static char rcsid[] = "@(#)$Id: outheaders.c,v 2.6 2012-06-16 18:22:44 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.6 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@siilo.FMI.FI> 
 *                  (was hurtta+elm@posti.FMI.FI, hurtta+elm@ozone.FMI.FI)
 *      or  Kari Hurtta <elm@elmme-mailer.org>
 *****************************************************************************/

#include "def_addr.h"
#include "hdr_imp.h"
#include "s_me.h"

#include <errno.h>
#ifndef ANSI_C
extern int errno;
#endif         

DEBUG_VAR(Debug,__FILE__,"header");


static unsigned char *s2us P_((char *str));
static unsigned char *s2us(str) 
     char *str;
{
    return (unsigned char *)str;
}

static const unsigned char *cs2us P_((const char *str));
static const unsigned char *cs2us(str) 
     const char *str;
{
    return (const unsigned char *)str;
}




int add_group_phrase_(x,GRP) 
     struct expanded_address *x;
     const struct string *GRP;
{
    if (x->magic   != EXP_ADDR_magic) 
    	panic("HEADERS PANIC",__FILE__,__LINE__,"add_group_phrase_",
	      "Bad magic number",0);
    
    if (!x->addrs)
	x->addrs = new_empty_addr_list(0);


    return add_group_to_list(x->addrs, GRP);
}

int add_expanded_addr0_ (x,ADR,g)
     struct expanded_address *x;
     const struct address *ADR;
     int g;
{
    if (x->magic   != EXP_ADDR_magic) 
    	panic("HEADERS PANIC",__FILE__,__LINE__,"add_expanded_addr0_",
	      "Bad magic number",0);

    if (!x->addrs)
	x->addrs = new_empty_addr_list(1);

    return add_address_to_list(x->addrs,ADR,g);
}

extern int add_expanded_addr_g(x,ADR,GRP)
     struct expanded_address *x;
     const struct address *ADR;
     const struct string *GRP;
{
    int g = -1;

    if (x->magic   != EXP_ADDR_magic) 
    	panic("HEADERS PANIC",__FILE__,__LINE__,"add_expanded_addr_g",
	      "Bad magic number",0);

    if (!x->addrs)
	x->addrs = new_empty_addr_list(1);

    if (GRP) 
	g = addr_list_lookup_group(x->addrs,GRP,1);
    
    return add_address_to_list(x->addrs,ADR,g);
}


int add_expanded_addr_ (x,ADR,FN,COM)
     struct expanded_address *x;
     const char *ADR;
     const struct string *FN;
     const struct string *COM;
{
    int X;

    struct address * a0 = new_address(ADR,FN,COM);

    X = add_expanded_addr0_(x,a0,-1);

    free_address(&a0);

    return X;
}

int add_expanded_addr_4(x,ADR,FN,COM,group)
     struct expanded_address *x;
     const char *ADR;
     const struct string *FN;
     const struct string *COM;
     const struct string *group;
{
    int X;

    struct address * a0 = new_address(ADR,FN,COM);

    X = add_expanded_addr_g(x,a0,group);

    free_address(&a0);

    return X;
}

int add_textual_addr_ (x,TEXTUAL,POS,LEN)
     struct expanded_address *x;
     const struct string *TEXTUAL;
     int POS, LEN;
{
    if (x->magic   != EXP_ADDR_magic) 
    	panic("HEADERS PANIC",__FILE__,__LINE__,"add_textual_addr_",
	      "Bad magic number",0);

    x->surface = safe_realloc(x->surface,(x->surface_len + 2) * 
			      sizeof (struct textual));
    if (!TEXTUAL)
	x->surface[x->surface_len].Textual     = NULL;
    else
	x->surface[x->surface_len].Textual     = dup_string(TEXTUAL);
    x->surface[x->surface_len].pos         = POS;
    x->surface[x->surface_len].len         = LEN;
    x->surface[x->surface_len+1].Textual   = NULL;
    x->surface[x->surface_len+1].pos       = POS+LEN;
    x->surface[x->surface_len+1].len       = 0;
    return x->surface_len++;
}

void zero_expanded_address (x)
     struct expanded_address *x;
{
    /* bzero is defined hdrs/elm_defs.h */
    bzero((void *)x,sizeof (*x));
    x->magic        = EXP_ADDR_magic;

    x->addrs        = NULL;
    x->surface      = NULL;
    x->surface_len  = 0;
}

void copy_expanded_address (result,source) 
     struct expanded_address *result;
     struct expanded_address source;
{
    int i;
    free_expanded_address(result);

    if (source.magic   != EXP_ADDR_magic) 
    	panic("HEADERS PANIC",__FILE__,__LINE__,"copy_expanded_address",
	      "Bad magic number",0);
    
    if (source.addrs) {
	int g, glen = addr_list_group_count(source.addrs);
	
	if (!result->addrs)
	    result->addrs = 
		new_empty_addr_list(addr_list_item_count(source.addrs));
	
	for (g = 0; g < glen; g++)
	    add_group_phrase_(result,addr_list_get_group(source.addrs,g));

    }

    for (i = 0; i < source.surface_len; i++) {
	int count = 0;
	int pos = result->addrs ? addr_list_item_count(result->addrs) : 0;
	int j;
	for (j = 0; j < source.surface[i].len; j++) {
	    int g;

	    const struct address * adr = addr_list_get_item(source.addrs,
							    source.surface[i].
							    pos+j,
							    &g);
	    int p = add_expanded_addr0_(result,adr,g);
	    
	    if (0 == count)
		pos = p;
	    count++;
	}
	add_textual_addr_(result,source.surface[i].Textual,pos,count);      
    }
}

void free_expanded_address(x)
     struct expanded_address *x;
{
    int i;
    
    if (x->magic   != EXP_ADDR_magic) 
    	panic("HEADERS PANIC",__FILE__,__LINE__,"free_expanded_address",
	      "Bad magic number",0);

    if (x->addrs) 
	free_addr_list(& x->addrs);
    
    if (x->surface) {
	for (i = 0; i < x->surface_len; i++) {
	    if (x->surface[i].Textual) {
		free_string (&(x->surface[i].Textual));	     
	    }
	}
	free(x->surface);
	x->surface = NULL;
    }
    x->surface_len = 0;
}

struct string * make_surface_addr(address)
     const struct address * address;
{
    struct string * result = new_string(display_charset);
    int fl = 0, cl = 0;

    const char          * addr     = 
	address_get_ascii_addr(address);
    const struct string * fullname = address_get_phrase(address);
    const struct string * comment  = address_get_comment(address);

    if (fullname &&
	(fl = string_len(fullname))) {

	int i;
	
	add_ascii_to_string(result,s2us("\""));
	
	for (i = 0; i < fl; i++) {
	    uint16 code = give_unicode_from_string(fullname,i);

	    /* No fancy quotation stuff on textual form for editing */
	    
	    switch(code) {
	    case 0x0022  /* '"'  */: 
	    case 0x005C  /* '\\' */: 
		code = 0x005F /* '_' */;
		break;
	    }
	    add_unicode_to_string(result,1,&code);
	}

       	add_ascii_to_string(result,s2us("\" "));
    }

    add_ascii_to_string(result,s2us("<"));
    if (addr)
	add_ascii_to_string(result,cs2us(addr));
    add_ascii_to_string(result,s2us(">"));
    
    if (comment && 
	(cl = string_len(comment))) {
	int i;

	add_ascii_to_string(result,s2us(" ("));

	for (i = 0; i < cl; i++) {
	    uint16 code = give_unicode_from_string(comment,i);

	    /* No fancy quotation stuff on textual form for editing */

	    switch(code) {
	    case 0x0022 /* '"'  */: 
	    case 0x005C /* '\\' */: 
	    case 0x0028 /* '('  */: 
	    case 0x0029 /* ')'  */: 
	    case 0x003A /* ':'  */:
	    case 0x003C /* '<'  */: 
	    case 0x003E /* '>'  */:
		code = '_';
		break;
	    }
	    add_unicode_to_string(result,1,&code);
	}

	add_ascii_to_string(result,s2us(")"));

    }

    return result;
}

void expanded_address_from_list(result,addr_list)
     struct expanded_address *result;
     const struct addr_list *addr_list;
{
    int j;
    int len,glen;
    int group_offset = 0;

    if (!addr_list)
	return;

    len =  addr_list_item_count(addr_list);
    glen = addr_list_group_count(addr_list);

    if (result->addrs)
	group_offset = addr_list_group_count(result->addrs);

    for (j = 0; j < glen; j++) {
	add_group_phrase_(result,
			  addr_list_get_group(addr_list,j));
    }

    for (j = 0; j < len; j++) {
	int g;
	int pos;
	struct string *s;

	const struct address * address =
	    addr_list_get_item(addr_list,j,&g);

	if (g >= 0)
	    g += group_offset;

	pos = add_expanded_addr0_(result,address,g);

	s = make_surface_addr(address); 

	add_textual_addr_(result,s,pos,1);
	free_string(&s);

    }
}


#ifdef ANSI_C
hdr_add_to_mailing_hdr add_from_hdr;
#endif
int add_from_hdr(headers,X,value,demime,defcharset,replace)
     struct mailing_headers * headers;
     header_ptr X;
     const char *value;
     int demime;
     charset_t defcharset;
     int replace;
{
    struct addr_list * list = 
	parse_header_address(NULL,value,demime,defcharset,NULL);

    if (!list)
	return 0;

    if (replace)
	free_expanded_address(&(headers->from));

    expanded_address_from_list(&(headers->from), list);

    free_addr_list(&list);

    return 1;
}


#ifdef ANSI_C
hdr_add_to_mailing_hdr add_to_hdr;
#endif
int add_to_hdr(headers,X,value,demime,defcharset,replace)
     struct mailing_headers * headers;
     header_ptr X;
     const char *value;
     int demime;
     charset_t defcharset;
     int replace;
{    
    struct addr_list * list = 
	parse_header_address(NULL,value,demime,defcharset,NULL);

    if (!list)
	return 0;

    if (replace)
	free_expanded_address(&(headers->to));

    expanded_address_from_list(&(headers->to), list);

    free_addr_list(&list);

    return 1;
}


#ifdef ANSI_C
hdr_add_to_mailing_hdr add_cc_hdr;
#endif
int add_cc_hdr(headers,X,value,demime,defcharset,replace)
     struct mailing_headers * headers;
     header_ptr X;
     const char *value;
     int demime;
     charset_t defcharset;
     int replace;
{
    struct addr_list * list = 
	parse_header_address(NULL,value,demime,defcharset,NULL);
    
    if (!list)
	return 0;

    if (replace)
	free_expanded_address(&(headers->cc));

    expanded_address_from_list(&(headers->cc), list);

    free_addr_list(&list);

    return 1;
}


#ifdef ANSI_C
hdr_add_to_mailing_hdr add_bcc_hdr;
#endif
int add_bcc_hdr(headers,X,value,demime,defcharset,replace)
     struct mailing_headers * headers;
     header_ptr X;
     const char *value;
     int demime;
     charset_t defcharset;
     int replace;
{
    struct addr_list * list = 
	parse_header_address(NULL,value,demime,defcharset,NULL);
    
    if (!list)
	return 0;

    if (replace)
	free_expanded_address(&(headers->bcc));

    expanded_address_from_list(&(headers->bcc), list);

    free_addr_list(&list);

    return 1;
}



#ifdef ANSI_C
hdr_add_to_mailing_hdr add_replyto_hdr;
#endif
int add_replyto_hdr(headers,X,value,demime,defcharset,replace)
     struct mailing_headers * headers;
     header_ptr X;
     const char *value;
     int demime;
     charset_t defcharset;
     int replace;
{
    struct addr_list * list = 
	parse_header_address(NULL,value,demime,defcharset,NULL);
    
    if (!list)
	return 0;

    if (replace)
	free_expanded_address(&(headers->reply_to));

    expanded_address_from_list(&(headers->reply_to), list);

    free_addr_list(&list);

    return 1;
}

#if ANSI_C
static env_from_change_hook dummy_env_from_change;
#endif
static int dummy_env_from_change P_((struct mailer_env_from *X,
				     const char * value));

static int dummy_env_from_change(X,value)
     struct mailer_env_from *X;
     const char * value;
{
    return 0;

}

static env_from_change_hook  * ef_hook = &dummy_env_from_change;

void set_env_from_change_hook(hook)
     env_from_change_hook *hook;
{
    ef_hook = hook;
}

#ifdef ANSI_C
hdr_add_to_mailing_hdr add_env_from_hdr;
#endif
int add_env_from_hdr(headers,X,value,demime,defcharset,replace)
     struct mailing_headers * headers;
     header_ptr X;
     const char *value;
     int demime;
     charset_t defcharset;
     int replace;
{
    int ret = 0;
    
    if (0 == strcmp(value,"<>")) {
	ret = ef_hook(headers->env_from,value);
    } else {

	struct addr_list * TMP = 
	    parse_header_address(NULL,value,demime,defcharset,NULL);
	
	if (TMP) {
	    int len = addr_list_item_count(TMP);

	    if (1 == len) {
		int group = -1;
		const struct address * address =
		    addr_list_get_item(TMP,0,&group);
		const char *addr = 
		    address_get_ascii_addr(address);

		if (addr && headers->env_from) {
		    char * t = elm_message(FRM("<%s>"),addr);
		    ret = ef_hook(headers->env_from,t);	    
		    free(t);
		}
	    }
	    
	    free_addr_list(&TMP);
	}
    }

    return ret;
}


/* This replaces always */
#ifdef ANSI_C
hdr_add_to_mailing_hdr add_subject_hdr;
#endif

int add_subject_hdr(headers,X,value,demime,defcharset,replace)
     struct mailing_headers * headers;
     header_ptr X;
     const char *value;
     int demime;
     charset_t defcharset;
     int replace;
{
    if (headers->subject)
	free_string( & (headers->subject)  );

    headers->subject = 
	hdr_to_string(HDR_TEXT,value,defcharset,demime);

    return 1;
}



#ifdef ANSI_C
hdr_add_to_mailing_hdr add_user_hdr;
#endif

int add_user_hdr(headers,X,value,demime,defcharset,replace)
     struct mailing_headers * headers;
     header_ptr X;
     const char *value;
     int demime;
     charset_t defcharset;
     int replace;
{
    if (replace) {

	/* Only replace header with same name */

	int i;
	int count = 0;

	for (i = 0; i < headers->user_header_count; i++) {

	    if (X == headers->user_header[i].name) {

		if (headers->user_header[i].value)
		    free_string(&(headers->user_header[i].value));
		headers->user_header[i].name = NULL;
	    } else {
		headers->user_header[count] = headers->user_header[i];

		if (count != i) {
		    headers->user_header[i].value = NULL;
		    headers->user_header[i].name = NULL;
		}
		count++;
	    }

	}
	headers->user_header_count = count;
    }

    headers->user_header = safe_realloc(headers->user_header,
				     (headers->user_header_count +1) *
				     sizeof (headers->user_header[0]));
    headers->user_header[headers->user_header_count].name = X;
    headers->user_header[headers->user_header_count].value = 
	hdr_to_string(HDR_TEXT,value,defcharset,demime);
    headers->user_header_count++;

    return 1;
}

void dump_expanded_address(debuglevel,text,expanded)
     int debuglevel; 
     const char *text;
     struct expanded_address expanded;

{
    struct textual *ptr;
    int l = strlen(text);
    int LASTPOS = -1;
    int alen = expanded.addrs ?
	addr_list_item_count(expanded.addrs) : 0;
    int glen = expanded.addrs ?
	addr_list_group_count(expanded.addrs) : 0;

    
    DPRINT(Debug,debuglevel,
	   (&Debug, 
	    "%s: %d surfaces, %d addresses, %d groups\n",
	    text, expanded.surface_len, alen, glen));
 
    for (ptr = expanded.surface; 
	 ptr < expanded.surface + expanded.surface_len; 
	 ptr++) {
	int j;
	if (!ptr->Textual) {
	    DPRINT(Debug,debuglevel,
		   (&Debug, 
		    "%*s [%d] surface NULL! \t(%d addresses)\n",
		    l,"",ptr-expanded.surface,ptr->len));
	} else {
	    DPRINT(Debug,debuglevel,
		   (&Debug, 
		    "%*s [%d] surface=%S \t(%d addresses)\n",
		    l,"",ptr-expanded.surface,ptr->Textual,
		    ptr->len));
	}
	if (ptr->pos != LASTPOS + 1) {     
	    DPRINT(Debug,debuglevel,
		   (&Debug, 
		    "%*s [%d] ERROR!  address pos = %d != %d (lastpost +1)\n",
		    l,"",ptr-expanded.surface, ptr->pos, LASTPOS+1));			
	}
	for (j = 0; j < ptr->len; j++) {
	    if (ptr->pos+j < 0 || ptr->pos+j >= alen) {
		DPRINT(Debug,debuglevel,
		       (&Debug, 
			"%*s    %d [%d] ERROR, not range 0 - %d\n",
			l,"",j,ptr->pos,
			alen-1));
	    } else {

		int group = -1;
		const struct address * address = 
		    addr_list_get_item(expanded.addrs,ptr->pos+j,&group);
		
		const char          * addr     = 
		    address_get_ascii_addr(address);
		const struct string * fullname = address_get_phrase(address);
		const struct string * comment  = address_get_comment(address);

		DPRINT(Debug,debuglevel,
		       (&Debug, 			    
			"%*s    %d [%d] addr=%s",
			l,"",j,ptr->pos+j,
			addr ? addr : "<NULL>"));

		if (comment) {
		    DPRINT(Debug,debuglevel,
			   (&Debug,"\tfullname=%S",
			    fullname));
		} else {
		    DPRINT(Debug,debuglevel,
			   (&Debug,"\tfullname=<NULL>"));

		}

		if (comment) {
		    DPRINT(Debug,debuglevel,
			   (&Debug,"\tcomment=%S\n",
			    comment));
		} else {
		    DPRINT(Debug,debuglevel,
			   (&Debug,"\tcomment=<NULL>\n",
			    comment));
		}
		
		if (fullname) {
		    const char *A, *B;

		    A = get_string_MIME_name(fullname);
		    B = get_string_lang(fullname);
		    
		    DPRINT(Debug,debuglevel,
			   (&Debug, 			    
			    "%*s         fullname cs=%s lang=%s, ",
			    l,"",
			    A ? A : "<none>",
			    B ? B : "<none>"));
		}


		if (comment) {
		    const char *A, *B;

		    A = get_string_MIME_name(comment);
		    B = get_string_lang(comment);
		    
		    DPRINT(Debug,debuglevel,
			   (&Debug,"comment cs=%s lang=%s\n",
			    A ? A : "<none>",
			    B ? B : "<none>"));
		}
		
		DPRINT(Debug,debuglevel,
		       (&Debug, 			    
			"%*s         group=%d",l,"",group));
		if (group >= glen) {
		    DPRINT(Debug,debuglevel,(&Debug," ERROR not range\n"));
		} else if (group >= 0) {
		    const struct string * G = addr_list_get_group(expanded.addrs,group);
		
		    DPRINT(Debug,debuglevel,(&Debug,"=%S\n",G));
		} else {
		    DPRINT(Debug,debuglevel,(&Debug,"\n"));
		}

		
		LASTPOS = ptr->pos+j;
	    }
	}
    }
    DPRINT(Debug,debuglevel,
	   (&Debug, 
	    "%*s === end\n",l,""));
}


/* Returns either backquote replaced string or original
   if return value != buffer, value is malloced 
*/

static char * eval_backquote P_((char *buffer));
static char * eval_backquote(buffer)
     char *buffer;
{
    char *ptr ;
    char *result  = NULL;
    char *x1      = buffer;
    char *cptr    = NULL;
    int quitflag  = 0;

    DPRINT(Debug,8,(&Debug, "eval_backquote: %s\n",
		    buffer));


    for (ptr = buffer; (*ptr || result) && !quitflag; ptr++) {

	if ('`' == *ptr || !*ptr) {
	    
	    char * start = cptr ? cptr : x1;
	    int len;
	    char * clip;

	    quitflag = !*ptr;

	    if (!start)
		panic("HEADERS PANIC",__FILE__,__LINE__,"eval_backquote",
		      "No start pointer",0);
	    
	    len = ptr - start;
	    if (len < 0)
		panic("HEADERS PANIC",__FILE__,__LINE__,"eval_backquote",
		      "Bad start pointer",0);
		
	    clip = safe_malloc(len+1);
	    memcpy(clip,start,len);
	    clip[len] = '\0';

	    DPRINT(Debug,11,(&Debug, "eval_backquote: clip(len=%d)=%s\n",
			     len,clip));

	    if (!cptr) {
		/* Got start of command */

		result = strmcat(result,clip);
		cptr = ptr +1;
		x1 = NULL;
		free(clip);
		continue;
	    }

	    if (len > 0) {
		struct run_state RS;
		const char * argv[4];
		int retcode;

		char local_buffer[80];
		const char * sh = give_dt_estr_as_str(&shell_e,"shell",
						      NULL,NULL);
		int n;

		int res;

		if (!sh)
		    sh = "/bin/sh";

		DPRINT(Debug,5,(&Debug, "eval_backquote: command: %s\n",
				clip));

		argv[0] = sh;
		argv[1] = "-c";
		argv[2] = clip;
		argv[3] = NULL;
		
		res = start_run(&RS,
				SY_ENAB_SIGHUP|SY_ENAB_SIGINT|
				SY_RUN_STATE_OPIPE,argv,-1,-1);
		if (0 == res) {
		    lib_error(CATGETS(elm_msg_cat, MeSet, 
				      MeBackquoteCmdFailed,
				      "Backquoted command \"%s\" in elmheaders failed."),
			      clip);
		    free(clip);
		    free(result);

		    DPRINT(Debug,5,(&Debug, 
				    "eval_backquote: start_run failed\n"));
		    return NULL;
		}

		while (0 < (n = 
			    mail_gets(local_buffer, sizeof local_buffer, 
				      RS.pfd))) {	
		    DPRINT(Debug,5,(&Debug, 
				    "eval_backquote: output: %s",
				    local_buffer));
		    if (local_buffer[n-1] != '\n') {
			DPRINT(Debug,5,(&Debug, 
				    "<-- no NEWLINE\n"));
		    }

		    result = strmcat(result,local_buffer);
		    
		}

		if (result && 0 < (n = strlen(result)) &&
		    '\n' == result[n-1]) {
		    result[n-1] = '\0';
		    DPRINT(Debug,5,(&Debug, 
				    "... Removed newline from end of result\n"));
		}

		res = wait_end(&RS,&retcode);
		if (res < 0 || retcode != 0) {
		    lib_error(CATGETS(elm_msg_cat, MeSet, 
				      MeBackquoteCmdFailed,
				      "Backquoted command \"%s\" in elmheaders failed."),
			      clip);
		    free(clip);
		    free(result);
		    
		    DPRINT(Debug,5,(&Debug, 
				    "eval_backquote: command failed: res=%d, retcode=%d\n",
				    res,retcode));
		    return NULL;

		}
	    }

	    x1 = ptr +1;
	    cptr = NULL;
	    free(clip);
	}
    }

    if (result) {
	DPRINT(Debug,8,(&Debug, "eval_backquote: result=%s\n",
			result));
    }
	
    return result ? result : buffer;
}


void import_mailheaders(headers)
     struct mailing_headers * headers;
{
    int err;
    FILE *F;
    char buffer[32*1024];

    if (0 != (err = can_open(user_mailheaders,"r"))) {
	DPRINT(Debug,5,(&Debug, "%s not available: %s (errno=%d)\n",
			user_mailheaders,strerror(err),err));
	return;
    }

    F = fopen(user_mailheaders,"r");
    if (!F) {
	/* In than point opening should have succeed */
	int err = errno;
	lib_error(FRM("%s: %s"),user_mailheaders,err);
	
	return;
    }

    while (0 < read_header_line(F,buffer,sizeof buffer,0)) {

	header_ptr header_name;

	char *k, *k1;
	char * value1 = NULL;

	value1 = eval_backquote(buffer);

	if (!value1)
	    continue;

	k = strchr(value1,':');
	if (!k)
	    break;

	/* Strip whitespace before ':' */
	k1 = k;
	while(k1 > value1 && 
	      whitespace(*(k1-1))) {
	    k1 --;
	    *k1 = '\0';
	}

	*k = '\0';
	k++;

	/* Skip whitespace after ':' */
	while (whitespace(*k))
	    k++;
	
	header_name = find_header(value1,1);
	
	if (!add_to_mailing_header(headers,header_name,k,
				   !allow_no_hdrencoding,
				   system_charset,0)) {
	    /* TODO: Notice failure ... */
	}

	if (value1 != buffer)
	    free(value1);

    }

    fclose(F);
}

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
